import os
import platform
import subprocess
import time
import shutil
import hashlib
import multiprocessing as mp
from datetime import datetime

from .config import *


# ---just a time formatter---
def ftime(ti):
    return time.strftime('%H:%M:%S', time.gmtime(ti))
# ---------

class Sampler:
    def __init__(self, current_os):
        self.current_os = current_os
        print("BUILDING FOR: ", self.current_os, platform.release())
        self.all_sample = None
        self.thread_number = None
        self.reboot_counter = 0
        self.start_time = None
        self.last_round_time = 0

        self.exec_name = defaults.get(self.current_os).get("exec_name")
        self.monitor_freq = defaults.get(self.current_os).get("monitor_freq")

        self.working_dir = defaults.get(self.current_os).get("dir_name")  # eg ./data_linux
        self.data_dir = self.working_dir+DATA_PATH  # eg ./data_linux/data/
        self.archive_dir = self.working_dir+ARCHIVE_PATH  # eg ./data_linux/archives/
        self.exec_path = EXECUTABLES_PATH + "/" + self.exec_name

        # check directory
        if not os.path.isdir("./"+self.working_dir):
            os.mkdir("./"+self.working_dir)
        if not os.path.isdir("./"+self.data_dir):
            os.mkdir("./"+self.data_dir)
        if not os.path.isdir("./"+self.archive_dir):
            os.mkdir("./"+self.archive_dir)

    def ask_initial_param(self):
        sam = defaults.get(self.current_os).get("n_samples")
        thr = defaults.get(self.current_os).get("n_threads")
        self.all_sample = int(input(f"Number of samples: ({sam})") or sam)
        self.thread_number = int(input(f"Number of threads: ({thr})") or thr)

    def quick_init(self, all_sample, thread_number, reboot_counter):
        self.all_sample = all_sample
        self.thread_number = thread_number
        self.reboot_counter = reboot_counter

    def get_initial_param(self):
        print(f"\nRequested {self.all_sample} samples on {self.thread_number} threads\n")

    def __monitor__(self, taken_sample):
        total_taken = taken_sample * self.thread_number
        interval = time.time() - self.last_round_time
        elapsed_time = time.time() - self.start_time
        rate_sec_sample = elapsed_time / (total_taken + 1)
        remaining_time = (self.all_sample - taken_sample * self.thread_number) * rate_sec_sample
        rate_sample_sec = round((self.monitor_freq * self.thread_number) / interval, 2)
        avg_rate_sample_sec = round(1 / rate_sec_sample, 2)
        percent_done = round((total_taken / self.all_sample) * 100, 2)

        width, height = shutil.get_terminal_size()
        bar_len = width - (len(str(percent_done)) + 4)
        done = round(percent_done / (100 / bar_len))
        togo = bar_len - done
        done_str = '█' * int(done)
        togo_str = '_' * int(togo)
        ending = "\x1b[K" + "\033[F\x1b[K" * height

        print(f"{ending}"
              f"SAMPLING RUNNING\tTotal: {self.all_sample}\t N: {total_taken}\n\n"
              f"Pass: {ftime(elapsed_time)}"  # elapsed time
              f"\tLeft: {ftime(remaining_time)}"  # remaining time
              f"\t\tCurr: {rate_sample_sec} smpl/s"
              f"\tAvg: {avg_rate_sample_sec} smpl/s"
              f"\n\n"
              f'[{done_str}{togo_str}] {percent_done}%', end="")  # actual rate

    def __compress__(self):
        print("Compressing")

        end = time.time()
        total = end - self.start_time

        archive_name = "raw_" + str(self.reboot_counter)
        meta_name = "meta_" + str(self.reboot_counter)
        shutil.make_archive(format='zip', root_dir=self.data_dir, base_name=self.archive_dir + archive_name)

        print("Computing Checksum")
        f = open(self.archive_dir + archive_name + ".zip", "rb")
        hs = hashlib.md5(f.read()).hexdigest()
        f.close()
        print(hs)
        f = open(self.archive_dir + meta_name + ".txt", "w+")
        f.write(hs)
        f.write(f"\nReboot: {self.reboot_counter}")
        f.write(f"\nSamples: {self.all_sample},"
                f" Threads: {self.thread_number},"
                f" Time: {ftime(total)},"
                f" Rate: {round(self.all_sample / total, 2)},"
                f" \nINFO: {platform.uname()}\n"
                f" \nDATE: {datetime.now()}")
        f.close()
        print("\nDONE")

    def __sample__(self):
        all_th = []


        for i in range(1, self.thread_number):
            process = mp.Process(target=self.__experimental_sample__, args=(i,))
            all_th.append(process)

        all_th.append(mp.Process(target=self.__experimental_sample__, args=(0,), kwargs={"monit": 1}))

        print(f"\nStarting Sampling for {self.all_sample} samples on {self.thread_number} threads\n")

        self.start_time = time.time()

        for e in all_th:
            e.start()

        for e in all_th:
            e.join()

        end = time.time()
        total = end - self.start_time
        print(f"Total Time: {ftime(total)}, medium rate: {round(self.all_sample / total, 2)} smpl/s")

    def __experimental_sample__(self, n, monit=0):

        file = open(f"{self.data_dir}/raw_{n}.txt", "w+")
        res = ""
        for j in range(0, (self.all_sample // self.thread_number)+1):  # j param NULL is reserved for testing purpose
            out = subprocess\
                .Popen(["./" + self.exec_path, str(j), str(n)], stdout=subprocess.PIPE)\
                .communicate()[0].decode()
            if "error" in out:
                print("check error:", out)
                exit(-1)
            out = out.replace("\r\n", "\n")
            res += out
            if monit and j % self.monitor_freq == 0:
                self.__monitor__(j)
                self.last_round_time = time.time()

            if not j % 100000:
                file.write(res)
                res = ""
        file.write(res)

    def start_sampling(self):
        self.__sample__()
        self.__compress__()


# _________________________________________________________________________________________
